const mongoose = require('mongoose');
const logger = require('../src/logging');
const superagent = require('superagent');
const Schema = mongoose.Schema;
const workerModel = require('../models/worker');
const serviceModel = require('../models/services');

let model;

let JobsModelSchema = new Schema({
    service: String,
    param: Object,
    result: Object,
    started: Number,
    state: String
});

JobsModelSchema.statics.removeOldJobs = async () => {
    let services = await serviceModel.find({});

    for(let service of services){
        let jobs = await model.find({'service': service.name,  $or: [{state: 'running'},{state: 'pending'}]}).where('started').lt(Date.now() - service.timeout).exec();

        for(let job of jobs){
            job.state = 'open';
            await job.save();

            let worker = await workerModel.findOne({activeJob: job._id});

            if(worker){
                worker.activeJob = null;
                worker.active = false;
                await worker.save();
            }
        }
    }
};

JobsModelSchema.statics.process = async () => {

    let run_job = true;
    let run_worker = true;

    const jobs = await model.aggregate([
        {
            '$match': {
                'state': 'open'
            }
        }, {
            '$lookup': {
                'from': 'services',
                'localField': 'service',
                'foreignField': 'name',
                'as': 'service'
            }
        }, {
            '$project': {
                'param': 1,
                'result': 1,
                'started': 1,
                'state': 1,
                'service': 1,
                'service': {
                    '$arrayElemAt': [
                        '$service', 0
                    ]
                }
            }
        }, {
            '$lookup': {
                'from': 'workers',
                'localField': 'service.name',
                'foreignField': 'service',
                'as': 'service.worker'
            }
        }
    ]).exec();

    for(let job of jobs){

        if(run_job && job.service.worker.length > 0){

            run_job = false;

            for(let worker of job.service.worker){
                if(worker.active === false && run_worker){

                    run_worker = false;

                    await workerModel.updateOne({ _id: worker._id }, { $set: { active: true, activeJob: job._id } });
                    await model.updateOne({ _id: job._id }, { $set: { state: "pending", started: Date.now() } });

                    superagent.post('http://' + worker.ip+ ':' +worker.port+ '/job')
                        .send({ id: job._id, param: job.param })
                        .set('Accept', 'application/json')
                        .then(async (res) => {
                            await model.updateOne({ _id: job._id }, { $set: { state: "running" } });
                        }).catch(async (err) => {
                            if(err) logger.error(err);

                            let up = await model.updateOne({ _id: job._id }, { $set: { state: "open", started: null } });

                            if(up){
                                await workerModel.removeWorker(worker._id);
                            }
                        });
                }
            }
        }
    }
};

JobsModelSchema.statics.getQueuePosition = async (id, call) => {
    let job = await model.findOne({_id: id});
    let count = await model.countDocuments({ _id: { $lt: job._id}, service: job.service, $or: [{state: 'open'}, {state: 'running'}] });
    await call(count);
};

JobsModelSchema.statics.success = async (id, result, call) => {

    let job = await model.findOne({_id: id});

    if(job){
        job.state = 'success';
        job.result = result;

        await job.save();

        let worker = await workerModel.findOne({'activeJob': id});

        if(worker){
            worker.active = false;
            worker.activeJob = null;

            let work = await worker.save();

            if(work){
                call();
            }
        }
    }
};

JobsModelSchema.statics.update = async (id, result, call) => {
    let job = await model.findOne({_id: id});

    job.result = result;
    let saved = await job.save();

    if(saved){
        call();
    }
};

model = mongoose.model('jobs', JobsModelSchema );
module.exports = model;
